import os
import scipy
import numpy as np
import scipy.spatial
from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor


class GraphConstructorSpatial(GraphConstructor):
    _estimator_type = "transformer"

    def __init__(self,
                 k_distance: int = 10,
                 atlas_name: str = 'ho',
                 atlas_folder: str = None,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for generating adjacency matrices
        from connectivity matrices. Selects the k nearest
        neighbours for each node based on spatial distance
        of the coordinates in the chosen atlas. This method
        can be applied to both DTI and rs-fMRI data. A user
        can construct their own atlas using defined spatial
        coordinates that match those regions used in the
        correspoding connectivity matrices.
        Adapted from Ktena et al, 2017.

        !!! danger
            Currently considered untested!
            See <a href='https://github.com/wwu-mmll/photonai_graph/issues/65' target='_blank'>Ticket</a>

        Parameters
        ----------
        k_distance: int
            the k nearest neighbours value, for the kNN algorithm.
        atlas_name: str,default="ho"
            name of the atlas coordinate file
        atlas_folder: str,default="ho"
            path to the atlas coordinate file
        one_hot_nodes: int,default=0
            Whether to generate a one hot encoding of the nodes in the matrix (1) or not (0)
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int,default=0
            whether to perform a fisher transform of each matrix (1) or not (0)
        use_abs_fisher: int,default=0
            changes the values to absolute values. Is applied after fisher transform and before z-score transformation
        zscore: int,default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
        use_abs_zscore: int,default=0
            whether to use the absolute values of the z-score transformation or allow for negative values
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data

        Example
        -------
        Use outside of a PHOTON pipeline

        ```python
        constructor = GraphConstructorSpatial(k_distance=7,
                                              atlas_name="ho_coords.csv",
                                              atlas_path="path/to/your/data/",
                                              fisher_transform=1,
                                              use_abs=1)
        ```

        Or as part of a pipeline

        ```python
        my_pipe.add(PipelineElement('GraphConstructorSpatial',
                                    hyperparameters={'k_distance': 7,
                                    'atlas_name': "ho_coords.csv", 'atlas_path': "path/to/your/data/"}))
        ```
       """
        super(GraphConstructorSpatial, self).__init__(one_hot_nodes=one_hot_nodes,
                                                      use_abs=use_abs,
                                                      fisher_transform=fisher_transform,
                                                      use_abs_fisher=use_abs_fisher,
                                                      zscore=zscore,
                                                      use_abs_zscore=use_abs_zscore,
                                                      adjacency_axis=adjacency_axis,
                                                      logs=logs)
        self.k_distance = k_distance
        self.atlas_name = atlas_name
        self.atlas_folder = atlas_folder

    def get_spatial(self, X):
        """Returns the adjacency based on the spatial matrix"""
        # get atlas coords
        coords = self.get_atlas_coords(atlas_name=self.atlas_name, root_folder=self.atlas_folder)
        # generate adjacency matrix
        dist, idx = self.distance_scipy_spatial(coords, k=self.k_distance, metric='euclidean')
        adjacency = self.adjacency(dist, idx).astype(np.float32)
        # turn adjacency into numpy matrix for concatenation
        adjacency = adjacency.toarray()
        # repeat into desired length
        adjacency = np.repeat(adjacency[np.newaxis, :, :, np.newaxis], X.shape[0], axis=0)

        return adjacency

    def transform(self, X) -> np.ndarray:
        """Transform input matrices accordingly"""
        adj, feat = self.get_mtrx(X)
        # do preparatory matrix transformations
        adj = self.prep_mtrx(adj)
        # threshold matrix
        adj = self.get_spatial(adj)
        # get feature matrix
        transformed = self.get_features(adj, feat)

        return transformed

    @staticmethod
    def distance_scipy_spatial(z, k, metric='euclidean'):
        """Compute exact pairwise distances."""
        d = scipy.spatial.distance.pdist(z, metric)
        d = scipy.spatial.distance.squareform(d)
        # k-NN photonai_graph.
        idx = np.argsort(d)[:, 1:k + 1]
        d.sort()
        d = d[:, 1:k + 1]

        return d, idx

    @staticmethod
    def get_atlas_coords(atlas_name, root_folder):
        """
            atlas_name   : name of the atlas used
        returns:
            matrix       : matrix of roi 3D coordinates in MNI space (num_rois x 3)
        """
        if atlas_name == 'ho':
            folder = os.path.dirname(os.path.abspath(__file__))
            coords_file = os.path.join(folder, 'ho_coords.csv')
            coords = np.genfromtxt(coords_file, delimiter=',')
            coords = np.delete(coords, 82, axis=0)

        elif atlas_name != 'ho' and root_folder is None:
            raise NotImplementedError('Automatic detection of atlases from photonai-neuro not implemented yet.'
                                      'Please wait for futures releases.')
        else:
            coords_file = os.path.join(root_folder, atlas_name + '_coords.csv')
            coords = np.loadtxt(coords_file, delimiter=',')

        return coords



import numpy as np
from scipy.stats import zscore
import warnings


def cofluct(X, quantiles: tuple = (0, 1), return_mat=True):
    """
    Computes cofluctuation time-series (per edge) for a nodes x timepoints matrix X.
    Based on https://www.pnas.org/content/early/2020/10/21/2005531117

    Parameters
    ----------
    quantiles: tuple,default=(0,1)
        list of lowest/highest quantile of edge events to use
        [0, 1]: all events = pearson corr; [0, .05]: bottom 5% of events; [.95, 1]: top 5% of events
    return_mat: bool,default=True
        Whether to return a connectivity matrix (True) or dictionary (False). The dict edge contains cofluctuation
        time-series (pairs_of_nodes x timepoints) and event timeseries.


    Returns
    -------
    float
        edge cofluctuation time-series dict (pairs_of_nodes x timepoints) and event timeseries as dict

    """
    X = np.float32(X)
    # get z-value time series (mean 0, std 1)
    z_ts = zscore(X, axis=1)

    # multiply z-scored time series to get edge time series (co-fluctuation)
    edge_ids = np.triu_indices(X.shape[0], k=1)
    edge_ts = np.einsum('ik,jk->ijk', z_ts, z_ts)[edge_ids]     # get upper triangular values

    # get event_ts as Root Mean Square over nodes
    event_ts = np.sqrt(np.mean(np.square(edge_ts), axis=0))

    # get correlations for quantile intervals of events
    q_low = np.quantile(np.abs(event_ts), q=quantiles[0])
    q_high = np.quantile(np.abs(event_ts), q=quantiles[1])
    edge_ts = edge_ts[:, (event_ts >= q_low) & (event_ts <= q_high)]

    # compute correlation/connectivity vector or matrix
    cors = np.mean(edge_ts, axis=1)
    if return_mat:
        tmp = np.eye(X.shape[0])
        tmp[np.triu_indices(tmp.shape[0], k=1)] = cors
        cors = tmp + tmp.T
        np.fill_diagonal(cors, 1)

    # return as dict
    else:
        warnings.warn("dicts can not be handled by photonai."
                      "Dicts are implemented for use outside of Photon.")
        edge_ids_str = [str(m) + '_' + str(n) for m, n in zip(edge_ids[0], edge_ids[1])]
        edge_ts_dict = {key: val for val, key in zip(edge_ts, edge_ids_str)}
        edge_ts_dict['event_ts'] = event_ts
        return edge_ts_dict

    return cors



import unittest
import numpy as np
import dgl
from photonai_graph.GraphConversions import dense_to_dgl


class DenseToDglTest(unittest.TestCase):

    def setUp(self):
        self.random4d = np.random.rand(20, 20, 20, 2)
        self.edge_dict = self.edge_dict = {(1, 0): 1, (2, 0): 1}

    def test_axis_argument(self):
        with self.assertRaises(NotImplementedError):
            dense_to_dgl(self.random4d)

    def test_matrix(self):
        g = dense_to_dgl(self.random4d, adjacency_axis=0, feature_axis=1)
        self.assertEqual(type(g[0]), dgl.DGLGraph)

    def test_nonsense_input(self):
        with self.assertRaises(ValueError):
            dense_to_dgl(self.edge_dict, adjacency_axis=0, feature_axis=1)

    def test_disconnected_graph(self):
        in_graph = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]])
        in_feats = np.array([[.9, .3, .2], [.8, .2, .1], [.7, .1, .0]])
        in_graph = np.expand_dims(in_graph, axis=(0, -1))
        in_feats = np.expand_dims(in_feats, axis=(0, -1))
        in_graph = np.concatenate((in_graph, in_feats), axis=-1)
        dgl_graph = dense_to_dgl(in_graph, adjacency_axis=0, feature_axis=1)[0]
        self.assertEqual(dgl_graph.num_nodes(), 3)
        self.assertEqual(dgl_graph.num_edges(), 2)
        self.assertTrue(np.array_equal(dgl_graph.ndata['feat'].numpy(), in_feats[0, ..., 0]))



import unittest
from scipy import sparse
import numpy as np
from photonai_graph.GraphConversions import sparse_to_dense


class SparseToDenseTest(unittest.TestCase):

    def setUp(self):
        # create sparse matrices
        sparse_matrix = sparse.eye(5)
        self.single_matrix = sparse_matrix
        self.check_features = np.random.rand(5, 5)
        self.features = sparse.csr_matrix(self.check_features.copy())
        self.matrix_list = [sparse_matrix] * 10
        self.feature_list = [self.features] * 10
        # create dense control matrix
        self.control_matrix = np.eye(5)
        self.edge_dict = {(1, 0): 1, (2, 0): 1}

    def test_single_no_feat(self):
        g = sparse_to_dense([self.single_matrix])
        self.assertTrue(np.array_equal(g[0], self.control_matrix))

    def test_list_no_feat(self):
        g = sparse_to_dense(self.matrix_list)
        for i in range(10):
            self.assertTrue(np.array_equal(g[i], self.control_matrix))

    def test_nonsene_no_feat(self):
        with self.assertRaises(Exception):
            sparse_to_dense(self.edge_dict)

    def test_single_feat(self):
        g = sparse_to_dense(self.single_matrix, features=self.features)
        self.assertEqual(np.shape(g), (5, 5, 2))
        self.assertTrue(np.array_equal(g[..., 0], self.control_matrix))
        self.assertTrue(np.array_equal(g[..., 1], self.check_features))

    def test_list_feat(self):
        g = sparse_to_dense(self.matrix_list, features=self.feature_list)
        self.assertEqual(np.shape(g), (10, 5, 5, 2))
        self.assertTrue(np.array_equal(g[0, ..., 1], self.check_features))

    def test_nonsense_feat(self):
        with self.assertRaises(Exception):
            sparse_to_dense(self.edge_dict, features=self.edge_dict)



from typing import List
from itertools import islice, combinations

import numpy as np

from photonai_graph.GraphConstruction.graph_constructor import GraphConstructor


class GraphConstructorRandomWalks(GraphConstructor):
    _estimator_type = "transformer"

    def __init__(self,
                 k_distance: int = 10,
                 number_of_walks: int = 10,
                 walk_length: int = 10,
                 window_size: int = 5,
                 no_edge_weight: int = 1,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for generating adjacency matrices
        from connectivity matrices. Generates a kNN matrix
        and performs random walks on these. The coocurrence
        of two nodes in those walks is then used to generate
        a higher-order adjacency matrix, by applying the kNN
        algorithm on the matrix again.
        Adapted from Ma et al., 2019.


        Parameters
        ----------
        k_distance: int
            the k nearest neighbours value, for the kNN algorithm.
        number_of_walks: int,default=10
            number of walks to take to sample the random walk matrix
        walk_length: int,default=10
            length of the random walk, as the number of steps
        window_size: int,default=5
            size of the sliding window from which to sample to coocurrence of two nodes
        no_edge_weight: int,default=1
            whether to return an edge weight (0) or not (1)
        one_hot_nodes: int,default=0
            Whether to generate a one hot encoding of the nodes in the matrix (1) or not (0)
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int,default=0
            whether to perform a fisher transform of each matrix (1) or not (0)
        use_abs_fisher: int,default=0
            changes the values to absolute values. Is applied after fisher transform and before z-score transformation
        zscore: int,default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
        use_abs_zscore: int,default=0
            whether to use the absolute values of the z-score transformation or allow for negative values
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data


        Example
        -------
        Use outside of a PHOTON pipeline

        ```python
        constructor = GraphConstructorRandomWalks(k_distance=5,
                                                  number_of_walks=25,
                                                  fisher_transform=1,
                                                  use_abs=1)
        ```

        Or as part of a pipeline

        ```python
        my_pipe.add(PipelineElement('GraphConstructorRandomWalks',
                                    hyperparameters={'k_distance': 5,
                                    'number_of_walks': 25}))
        ```
       """
        super(GraphConstructorRandomWalks, self).__init__(one_hot_nodes=one_hot_nodes,
                                                          use_abs=use_abs,
                                                          fisher_transform=fisher_transform,
                                                          use_abs_fisher=use_abs_fisher,
                                                          zscore=zscore,
                                                          use_abs_zscore=use_abs_zscore,
                                                          adjacency_axis=adjacency_axis,
                                                          logs=logs)
        self.k_distance = k_distance
        self.number_of_walks = number_of_walks
        self.walk_length = walk_length
        self.window_size = window_size
        self.no_edge_weight = no_edge_weight

    @staticmethod
    def random_walk(adjacency: np.ndarray, walk_length: int, num_walks: int) -> List[List[List[int]]]:
        """Performs a random walk on a given photonai_graph"""
        # a -> adj
        # i -> starting row
        walks = []  # holds transitions
        elements = np.arange(adjacency.shape[0])  # for our photonai_graph [0,1,2,3]
        for _ in range(num_walks):
            node_walks = []
            for _ in range(elements.shape[0]):
                index = np.random.choice(elements, replace=False)  # current index for this iteration
                count = 0  # count of transitions
                walk = []
                while count < walk_length:
                    count += 1
                    walk.append(index)
                    probs = adjacency[index]  # probability of transitions
                    # sample from probs
                    sample = np.random.choice(elements, p=probs)  # sample a target using probs
                    index = sample  # go to target
                node_walks.append(walk)
            walks.append(node_walks)

        return walks

    @staticmethod
    def sliding_window(seq, n: int):
        """Returns a sliding window (of width n) over data from the iterable
            s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   """
        it = iter(seq)
        result = tuple(islice(it, n))
        if len(result) == n:
            yield result
        for elem in it:
            result = result[1:] + (elem,)
            yield result

    def sliding_window_frequency(self, x_mean: np.ndarray, walk_list: List[List[List[int]]]) -> np.ndarray:

        coocurrence = np.zeros((x_mean.shape[0], x_mean.shape[1]))

        for i in walk_list:
            node_walks = i
            for x in node_walks:
                for w in self.sliding_window(x, self.window_size):
                    for subset in combinations(w, 2):
                        coord1 = subset[0]
                        coord2 = subset[1]
                        coocurrence[coord1, coord2] += 1
                    # if they occur at the same time add + 1 to the frequency table

        return coocurrence

    def get_ho_adjacency(self, adj: np.ndarray) -> np.ndarray:
        """Returns the higher order adjacency of a graph"""
        adj = np.squeeze(adj)

        adjacency_list = self.__adjacency_to_list(adj, adj)
        ho_adjacency = adjacency_list[..., np.newaxis]

        return ho_adjacency

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Transforms the matrix using random walks"""
        adj, feat = self.get_mtrx(X)
        # do preparatory matrix transformations
        adj = self.prep_mtrx(adj)
        # threshold matrix
        adj = self.get_ho_adjacency(adj)
        # get feature matrix
        x_transformed = self.get_features(adj, feat)

        return x_transformed

    # !===== Helper Functions =====!
    def __adjacency_to_list(self, adjacency_in: np.ndarray, adjacency_rw: np.ndarray) -> np.ndarray:
        adjacency_list = []
        for i in range(adjacency_in.shape[0]):
            d, idx = self.distance_sklearn_metrics(adjacency_rw[i, :, :], k=self.k_distance, metric='euclidean')
            adjacency = self.adjacency(d, idx).astype(np.float32)

            # normalize adjacency
            if self.no_edge_weight == 1:
                adjacency[adjacency > 0] = 1

            adjacency = adjacency.toarray()
            higherorder_adjacency = self.__adjacency_to_dense(adjacency)
            # turn adjacency into numpy matrix for concatenation
            adjacency_list.append(higherorder_adjacency)

        return np.asarray(adjacency_list)

    def __adjacency_to_dense(self, adjacency: np.ndarray) -> np.ndarray:
        adjacency_rowsum = np.sum(adjacency, axis=1)
        adjacency_norm = adjacency / adjacency_rowsum[:, np.newaxis]

        walks = self.random_walk(adjacency=adjacency_norm, walk_length=self.walk_length,
                                 num_walks=self.number_of_walks)

        higherorder_adjacency = self.sliding_window_frequency(x_mean=adjacency, walk_list=walks)

        # obtain the kNN photonai_graph from the new adjacency matrix
        d, idx = self.distance_sklearn_metrics(higherorder_adjacency, k=10, metric='euclidean')
        higherorder_adjacency = self.adjacency(d, idx).astype(np.float32)

        # convert this adjacency matrix to dense format
        return higherorder_adjacency.toarray()



import unittest
import numpy as np
from photonai_graph.DynamicUtils.cofluct_transform import CofluctTransform


class CofluctTransformTests(unittest.TestCase):

    def setUp(self):
        self.X_test2d = np.random.rand(20, 2000)
        self.X_test3d = np.random.rand(20, 20, 2000)
        self.X_test4d = np.random.rand(20, 20, 2000, 2)
        self.y = np.random.rand(20)

    def test_3d(self):
        coflu = CofluctTransform(quantiles=(0, 0.1), return_mat=True)
        coflu.fit(self.X_test3d, self.y)
        vec = coflu.transform(self.X_test3d)
        self.assertEqual(np.shape(vec), (20, 20, 20))

    def test_4d(self):
        coflu = CofluctTransform(quantiles=(0, 1), return_mat=True)
        coflu.fit(self.X_test4d, self.y)
        vec = coflu.transform(self.X_test4d)
        self.assertEqual(np.shape(vec), (20, 20, 20))

    def test_2d_error(self):
        coflu = CofluctTransform(quantiles=(0, 1), return_mat=True)
        with self.assertRaises(TypeError):
            coflu.fit(self.X_test2d, self.y)
            coflu.transform(self.X_test2d)

    def test_return_dict(self):
        coflu = CofluctTransform(quantiles=(0, 1), return_mat=False)
        coflu.fit(self.X_test3d, self.y)
        vec = coflu.transform(self.X_test3d)
        self.assertEqual(type(vec[0]), dict)



import unittest
import numpy as np
import scipy.sparse
import networkx as nx
from photonai_graph.GraphUtilities import draw_connectivity_matrix


class DrawConnectivityMatrixTest(unittest.TestCase):

    def setUp(self):
        # numpy matrices
        self.m2d_dense = np.random.rand(20, 20)
        self.m3d_dense = np.random.rand(10, 20, 20)
        self.m3d_dense_adjacency = np.random.rand(20, 20, 2)
        self.m3d_dense_adjacency_list = [np.random.rand(20, 20, 2)] * 10
        self.m4d_dense = np.random.rand(10, 20, 20, 2)
        self.m5d_dense = np.random.rand(10, 20, 20, 2, 2)
        # sparse matrices
        self.m2d_sparse = [scipy.sparse.rand(20, 20)] * 5
        # networkx graphs
        self.nx_graph_single = nx.erdos_renyi_graph(20, p=0.3)
        self.nx_graphs = [nx.erdos_renyi_graph(20, p=0.3)] * 10

    def test_2d_dense(self):
        draw_connectivity_matrix(self.m2d_dense, show=False)

    def test_2d_colorbar(self):
        draw_connectivity_matrix(self.m2d_dense, colorbar=True, show=False)

    def test_2d_adjacency_axis(self):
        with self.assertRaises(Exception):
            draw_connectivity_matrix(self.m2d_dense, adjacency_axis=0, show=False)

    def test_3d_dense(self):
        draw_connectivity_matrix(self.m3d_dense, show=False)

    def test_3d_dense_colorbar(self):
        draw_connectivity_matrix(self.m3d_dense, colorbar=True, show=False)

    def test_3d_dense_adjacency(self):
        draw_connectivity_matrix(self.m3d_dense_adjacency, colorbar=True, adjacency_axis=0, show=False)

    def test_3d_dense_adjacency_list(self):
        draw_connectivity_matrix(self.m3d_dense_adjacency_list, show=False)

    def test_3d_dense_adjacency_list_specified(self):
        draw_connectivity_matrix(self.m3d_dense_adjacency_list, colorbar=True,
                                 adjacency_axis=0, show=False)

    def test_4d_dense(self):
        with self.assertRaises(Exception):
            draw_connectivity_matrix(self.m4d_dense, show=False)

    def test_4d_dense_adjacency(self):
        draw_connectivity_matrix(self.m4d_dense, adjacency_axis=0, show=False)

    def test_4d_dense_colorbar(self):
        draw_connectivity_matrix(self.m4d_dense, colorbar=True, adjacency_axis=0, show=False)

    def test_5d_dense_no_adjacency(self):
        with self.assertRaises(Exception):
            draw_connectivity_matrix(self.m5d_dense)

    def test_2d_sparse(self):
        draw_connectivity_matrix(self.m2d_sparse, show=False)

    def test_wrong_type(self):
        with self.assertRaises(TypeError):
            draw_connectivity_matrix(self.nx_graphs, show=False)

    def test_wrong_type_single(self):
        with self.assertRaises(TypeError):
            draw_connectivity_matrix(self.nx_graph_single, show=False)



import numpy as np
try:
    import dgl
    import torch.nn as nn
    import torch.nn.functional as F
    from dgl.nn.pytorch import GraphConv
except ImportError:
    pass

from photonai_graph.NeuralNets.dgl_base import DGLRegressorBaseModel, DGLClassifierBaseModel


class GCNClassifier(nn.Module):
    def __init__(self, in_dim, hidden_dim, n_classes, hidden_layers, allow_zero_in_degree):
        super(GCNClassifier, self).__init__()
        self.layers = nn.ModuleList()
        # input layers
        self.layers.append(GraphConv(in_dim, hidden_dim, allow_zero_in_degree=allow_zero_in_degree))
        # hidden layers
        for _ in range(1, hidden_layers):
            self.layers.append(GraphConv(hidden_dim, hidden_dim, allow_zero_in_degree=allow_zero_in_degree))
        self.classify = nn.Linear(hidden_dim, n_classes)

    def forward(self, g):
        # Use node degree as the initial node feature. For undirected graphs, the in-degree
        # is the same as the out_degree.
        h = g.in_degrees().view(-1, 1).float()
        # Perform graph convolution and activation function.
        for i, gnn in enumerate(self.layers):
            h = F.relu(gnn(g, h))
        g.ndata['h'] = h
        # Calculate graph representation by averaging all the node representations.
        hg = dgl.mean_nodes(g, 'h')
        return self.classify(hg)


class GCNClassifierModel(DGLClassifierBaseModel):

    def __init__(self,
                 in_dim: int = 1,
                 hidden_layers: int = 2,
                 hidden_dim: int = 256,
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 early_stopping: bool = False,
                 es_patience: int = 10,
                 es_tolerance: int = 9,
                 es_delta: float = 0,
                 verbose: bool = False,
                 logs: str = '',
                 **kwargs):
        """
        Graph Attention Network for graph classification. GCN Layers
        from Kipf & Welling, 2017.
        Implementation based on dgl & pytorch.


        Parameters
        ----------
        in_dim: int,default=1
            input dimension
        hidden_layers: int,default=2
            number of hidden layers used by the model
        hidden_dim: int,default=256
            dimensions in the hidden layers
        validation_score: bool,default=False
            If true the input data is split into train and test (90%/10%).
            The testset is then used to get validation results during training
        verbose: bool,default=False
            If true verbose output is generated
        """
        super(GCNClassifierModel, self).__init__(nn_epochs=nn_epochs,
                                                 learning_rate=learning_rate,
                                                 batch_size=batch_size,
                                                 adjacency_axis=adjacency_axis,
                                                 feature_axis=feature_axis,
                                                 add_self_loops=add_self_loops,
                                                 allow_zero_in_degree=allow_zero_in_degree,
                                                 validation_score=validation_score,
                                                 early_stopping=early_stopping,
                                                 es_patience=es_patience,
                                                 es_tolerance=es_tolerance,
                                                 es_delta=es_delta,
                                                 verbose=verbose,
                                                 logs=logs,
                                                 **kwargs)
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        self.hidden_layers = hidden_layers

    def _init_model(self, X=None, y=None):
        self.model = GCNClassifier(self.in_dim,
                                   self.hidden_dim,
                                   len(np.unique(y)),
                                   self.hidden_layers,
                                   allow_zero_in_degree=self.allow_zero_in_degree)


class GCNRegressorModel(DGLRegressorBaseModel):

    def __init__(self,
                 in_dim: int = 1,
                 hidden_layers: int = 2,
                 hidden_dim: int = 256,
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 early_stopping: bool = False,
                 es_patience: int = 10,
                 es_tolerance: int = 9,
                 es_delta: float = 0,
                 verbose: bool = False,
                 logs: str = '',
                 **kwargs):
        """
        Graph convolutional Network for graph regression. GCN Layers
        from Kipf & Welling, 2017. Implementation based on dgl & pytorch.


        Parameters
        ----------
        in_dim: int,default=1
            input dimension
        hidden_layers: int,default=2
            number of hidden layers used by the model
        hidden_dim: int,default=256
            dimensions in the hidden layers
        validation_score: bool,default=False
            If true the input data is split into train and test (90%/10%).
            The testset is then used to get validation results during training
        verbose: bool,default=False
            If true verbose output is generated
        """
        super(GCNRegressorModel, self).__init__(nn_epochs=nn_epochs,
                                                learning_rate=learning_rate,
                                                batch_size=batch_size,
                                                adjacency_axis=adjacency_axis,
                                                feature_axis=feature_axis,
                                                add_self_loops=add_self_loops,
                                                allow_zero_in_degree=allow_zero_in_degree,
                                                validation_score=validation_score,
                                                early_stopping=early_stopping,
                                                es_patience=es_patience,
                                                es_tolerance=es_tolerance,
                                                es_delta=es_delta,
                                                verbose=verbose,
                                                logs=logs,
                                                **kwargs)
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        self.hidden_layers = hidden_layers

    def _init_model(self, X=None, y=None):
        self.model = GCNClassifier(self.in_dim,
                                   self.hidden_dim, 1,
                                   self.hidden_layers,
                                   allow_zero_in_degree=self.allow_zero_in_degree).float()



import unittest

import networkx as nx
import numpy as np
from photonai_graph.GraphUtilities import draw_connectogram
from photonai_graph.GraphConversions import dense_to_networkx


class DrawConnectogramTests(unittest.TestCase):

    def setUp(self):
        self.cyc_graph = nx.cycle_graph(20)
        self.weight_graph = dense_to_networkx(np.random.rand(1, 20, 20, 2), adjacency_axis=0)

    def test_drawing(self):
        draw_connectogram(self.cyc_graph, show=False)

    def test_drawing_node_shape(self):
        draw_connectogram(self.cyc_graph, node_shape='x', show=False)

    def test_drawing_colorscheme(self):
        draw_connectogram(self.cyc_graph, colorscheme="Blues", show=False)

    def test_drawing_nodesize(self):
        draw_connectogram(self.cyc_graph, nodesize=200, show=False)

    def test_drawing_weight(self):
        with self.assertRaises(KeyError):
            draw_connectogram(self.cyc_graph, weight=0.5, show=False)

    def test_drawing_weight_corrected(self):
        draw_connectogram(self.weight_graph[0], weight=0.5, show=False)



import os
from abc import ABC, abstractmethod

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.metrics import pairwise_distances
from scipy import sparse

from photonai_graph.GraphUtilities import individual_ztransform, individual_fishertransform


class GraphConstructor(BaseEstimator, TransformerMixin, ABC):
    _estimator_type = "transformer"

    def __init__(self,
                 one_hot_nodes: int = 0,
                 use_abs: int = 0,
                 fisher_transform: int = 0,
                 discard_original_connectivity: bool = False,
                 use_abs_fisher: int = 0,
                 zscore: int = 0,
                 use_abs_zscore: int = 0,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Base class for all graph constructors. Implements
        methods shared by different constructors.

        Parameters
        ----------
        one_hot_nodes: int, default=0
            whether to return a one hot node encoding as feature or not
        use_abs: bool, default = False
            whether to convert all matrix values to absolute values before applying
            other transformations
        fisher_transform: int, default=0
            whether to perform a fisher transform of every matrix
        discard_original_connectivity: bool,default=False
            If true the second index of the last dimension will be the original connectivity.
            Otherwise the original connectivity will be dropped from the matrix.
        use_abs_fisher: int
            Changes the values to absolute values. Is applied after fisher transform and before
            z-score transformation
        zscore: int, default=0
            performs a zscore transformation of the data. Applied after fisher transform and np_abs
            eval_final_perfomance is set to True
        use_abs_zscore: int, default=0
            whether to use the absolute values of the z-score transformation or allow for negative
            values. Applied after fisher transform, use_abs and zscore
        adjacency_axis: int, default=0
            position of the adjacency matrix, default being zero
        logs: str, default=None
            Path to the log data
        """
        if one_hot_nodes not in [0, 1]:
            raise ValueError("one_hot_nodes must be in [0, 1]")
        self.one_hot_nodes = one_hot_nodes
        self.use_abs = use_abs
        self.fisher_transform = fisher_transform
        self.use_abs_fisher = use_abs_fisher
        self.discard_original_connectivity = discard_original_connectivity
        self.use_abs_fisher = use_abs_fisher
        self.zscore = zscore
        self.use_abs_zscore = use_abs_zscore
        self.adjacency_axis = adjacency_axis
        self.logs = logs
        if not self.logs:
            self.logs = os.getcwd()
        self.mean_matrix = None

    def fit(self, X, y):
        return self

    def get_mtrx(self, graph_obj: np.ndarray) -> (np.ndarray, np.ndarray):
        """Returns a feature and adjacency matrix"""

        if np.ndim(graph_obj) != 4:
            raise ValueError("Please make sure your graphs have the needed input shape.")

        adjacency_matrix = graph_obj[:, :, :, self.adjacency_axis].copy()
        adjacency_matrix = adjacency_matrix[:, :, :, np.newaxis]
        feature_matrix = graph_obj.copy()

        return adjacency_matrix, feature_matrix

    def prep_mtrx(self, adjacency: np.ndarray) -> np.ndarray:
        """transforms the matrix according to selected criteria"""
        if self.use_abs:
            adjacency = np.abs(adjacency)
        if self.fisher_transform:
            adjacency = individual_fishertransform(adjacency)
            if self.use_abs_fisher:
                adjacency = np.abs(adjacency)
        if self.zscore:
            adjacency = individual_ztransform(adjacency)
            if self.use_abs_zscore:
                adjacency = np.abs(adjacency)

        return adjacency

    def get_features(self, adjacency: np.ndarray, features: np.ndarray) -> np.ndarray:
        """Returns a concatenated version of feature matrix and one hot nodes"""
        if self.discard_original_connectivity:
            features = features[..., 1:]
        if self.one_hot_nodes:
            adjacency = self.get_one_hot_nodes(adjacency)
        transformed = np.concatenate((adjacency, features), axis=3)
        return transformed

    @staticmethod
    def get_one_hot_nodes(adjacency: np.ndarray) -> np.ndarray:
        """Returns a one hot node encoding for every node of the matrix"""
        # an identity matrix is a one hot node encoding
        identity_matrix = np.identity((adjacency.shape[1]))
        # expands dims
        identity_matrix = identity_matrix[np.newaxis, :, :, np.newaxis]
        # repeat along first dimension and concatenate
        one_hot = np.repeat(identity_matrix, adjacency.shape[0], 0)
        concat = np.concatenate((adjacency, one_hot), axis=3)
        return concat

    @staticmethod
    def adjacency(dist, idx):
        """Return the adjacency matrix of a kNN photonai_graph."""
        m, k = dist.shape
        assert m, k == idx.shape
        assert dist.min() >= 0

        # Weights.
        sigma2 = np.mean(dist[:, -1]) ** 2
        dist = np.exp(- dist ** 2 / sigma2)

        # Weight matrix.
        i = np.arange(0, m).repeat(k)
        j = idx.reshape(m * k)
        v = dist.reshape(m * k)
        w = sparse.coo_matrix((v, (i, j)), shape=(m, m))

        # No self-connections.
        w.setdiag(0)

        # Non-directed photonai_graph.
        bigger = w.T > w
        w = w - w.multiply(bigger) + w.T.multiply(bigger)

        assert w.nnz % 2 == 0
        assert np.abs(w - w.T).mean() < 1e-10
        assert type(w) is sparse.csr.csr_matrix

        return w

    @staticmethod
    def distance_sklearn_metrics(z: np.ndarray, k: int, metric: str = 'euclidean') -> (np.ndarray, np.ndarray):
        """Compute exact pairwise distances."""
        d = pairwise_distances(z, metric=metric, n_jobs=-2)
        # k-NN photonai_graph.
        idx = np.argsort(d)[:, 1:k + 1]
        d.sort()
        d = d[:, 1:k + 1]
        return d, idx

    @abstractmethod
    def transform(self, X) -> np.ndarray:
        """transform the input data"""



import unittest
import numpy as np
import scipy.sparse
from photonai_graph.GraphUtilities import get_random_connectivity_data


class DrawConnectogramTests(unittest.TestCase):

    def setUp(self):
        self.np_control = np.random.rand(20, 20, 20, 2)

    def test_matrix_type(self):
        mtrx = get_random_connectivity_data(out_type="dense", number_of_nodes=20,
                                            number_of_individuals=20, number_of_modalities=2)
        self.assertEqual(type(self.np_control), type(mtrx))

    def test_matrix_shape(self):
        mtrx = get_random_connectivity_data(out_type="dense", number_of_nodes=20,
                                            number_of_individuals=20, number_of_modalities=2)
        self.assertEqual(np.shape(self.np_control), np.shape(mtrx))

    def test_sparse_type(self):
        mtrx = get_random_connectivity_data(out_type="sparse", number_of_nodes=20,
                                            number_of_individuals=20, number_of_modalities=2)
        self.assertEqual(type(mtrx), list)

    def test_sparse_len_individuals(self):
        mtrx = get_random_connectivity_data(out_type="sparse", number_of_nodes=20,
                                            number_of_individuals=20, number_of_modalities=2)
        self.assertEqual(len(mtrx), 20)

    def test_sparse_len_modality(self):
        mtrx = get_random_connectivity_data(out_type="sparse", number_of_nodes=20,
                                            number_of_individuals=20, number_of_modalities=2)
        self.assertEqual(len(mtrx[0]), 2)

    def test_sparse_type_sparse(self):
        mtrx = get_random_connectivity_data(out_type="sparse", number_of_nodes=20,
                                            number_of_individuals=20, number_of_modalities=2)
        individual_one = mtrx[0]
        self.assertEqual(type(individual_one[0]), scipy.sparse.coo_matrix)

    def test_wrong_type(self):
        with self.assertRaises(NotImplementedError):
            get_random_connectivity_data(out_type="networkx", number_of_nodes=20,
                                         number_of_individuals=20, number_of_modalities=2)



from photonai_graph.GraphEmbedding.graph_embedding_base import GraphEmbeddingBase
try:
    from gem.embedding.lle import LocallyLinearEmbedding
    from gem.embedding.static_graph_embedding import StaticGraphEmbedding
except ImportError:  # pragma: no cover
    pass


class GraphEmbeddingLocallyLinearEmbedding(GraphEmbeddingBase):
    _estimator_type = "transformer"

    def __init__(self,
                 embedding_dimension: int = 1,
                 adjacency_axis: int = 0,
                 logs: str = None):
        """
        Transformer class for calculating a Graph Embedding
        based on Locally Linear Embedding (Roweis & Saul,
        2000).
        Implementation based on gem python package.


        Parameters
        ----------
        embedding_dimension: int,default=1
            the number of dimensions that the final embedding will have
        adjacency_axis: int,default=0
            position of the adjacency matrix, default being zero
        logs: str,default=None
            Path to the log data


        Example
        -------
        ```python
        constructor = GraphEmbeddingLocallyLinearEmbedding(embedding_dimension=1)
        ```
        """
        super(GraphEmbeddingLocallyLinearEmbedding, self).__init__(embedding_dimension=embedding_dimension,
                                                                   adjacency_axis=adjacency_axis,
                                                                   logs=logs)

    def _init_embedding(self) -> StaticGraphEmbedding:
        return LocallyLinearEmbedding(d=self.embedding_dimension)



from photonai_graph.GraphEmbedding.graph_embedding_base import GraphEmbeddingBase
from photonai_graph.GraphEmbedding.graph_embedding_hope import GraphEmbeddingHOPE
from photonai_graph.GraphEmbedding.graph_embedding_laplacian_eigenmaps import GraphEmbeddingLaplacianEigenmaps
from photonai_graph.GraphEmbedding.graph_embedding_locally_linear_embedding import GraphEmbeddingLocallyLinearEmbedding



import sys

import networkx as nx
from photonai.base import PhotonRegistry
import os
try:
    import dgl
except ImportError:  # pragma: no cover
    pass
try:
    import grakel
except ImportError:  # pragma: no cover
    pass
try:
    import torch
except ImportError:  # pragma: no cover
    pass
try:
    from gem.embedding.hope import HOPE
except ImportError:  # pragma: no cover
    pass
try:
    import euler
except ImportError:  # pragma: no cover
    pass

"""This file is by intention excluded from coverage calculation"""


def delete_photonai_graph():  # pragma: no cover
    reg = PhotonRegistry()
    reg.delete_module('photonai_graph')


def register_photonai_graph():  # pragma: no cover
    current_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "photonai_graph/photonai_graph.json")
    reg = PhotonRegistry()
    reg.add_module(current_path)


def assert_imported(packages: list = None):  # pragma: no cover
    for package in packages:
        if package in sys.modules:
            return True
        else:
            raise ImportError(f"Could not load {package}. Please install it according to the documentation")


class NetworkxGraphWrapper:
    def __init__(self, graph_to_wrap: nx.Graph = None):
        self.inner_graph = graph_to_wrap



import numpy as np
try:
    import dgl
    import torch.nn as nn
    from dgl.nn.pytorch import SGConv
except ImportError:
    pass

from photonai_graph.NeuralNets.dgl_base import DGLClassifierBaseModel, DGLRegressorBaseModel


class SGConvClassifier(nn.Module):
    def __init__(self, in_dim, hidden_dim, n_classes, hidden_layers, allow_zero_in_degree):
        super(SGConvClassifier, self).__init__()
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        self.n_classes = n_classes
        self.hidden_layers = hidden_layers
        self.layers = nn.ModuleList()

        # input layer
        self.layers.append(SGConv(in_dim, hidden_dim))
        # hidden layers
        for _ in range(1, hidden_layers):
            self.layers.append(SGConv(hidden_dim, hidden_dim, allow_zero_in_degree=allow_zero_in_degree))
        # output layer
        self.classify = nn.Linear(hidden_dim, n_classes)

    def forward(self, bg):
        h = bg.in_degrees().view(-1, 1).float()
        for lr, layer in enumerate(self.layers):
            h = layer(bg, h)
        bg.ndata['h'] = h
        # Calculate graph representation by averaging all the node representations.
        hg = dgl.mean_nodes(bg, 'h')

        return self.classify(hg)


class SGConvClassifierModel(DGLClassifierBaseModel):

    def __init__(self,
                 in_dim: int = 1,
                 hidden_layers: int = 2,
                 hidden_dim: int = 256,
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 early_stopping: bool = False,
                 es_patience: int = 10,
                 es_tolerance: int = 9,
                 es_delta: float = 0,
                 verbose: bool = False,
                 logs: str = '',
                 **kwargs):
        """
        Graph convolutional network for graph classification. Simple Graph
        convolutional layers from Wu, Felix, et al., 2018.
        Implementation based on dgl & pytorch.


        Parameters
        ----------
        in_dim: int,default=1
            input dimension
        hidden_layers: int,default=2
            number of hidden layers used by the model
        hidden_dim: int,default=256
            dimensions in the hidden layers
        validation_score: bool,default=False
            If true the input data is split into train and test (90%/10%).
            The testset is then used to get validation results during training
        verbose: bool,default=False
            If true verbose output is generated
        """
        super(SGConvClassifierModel, self).__init__(nn_epochs=nn_epochs,
                                                    learning_rate=learning_rate,
                                                    batch_size=batch_size,
                                                    adjacency_axis=adjacency_axis,
                                                    feature_axis=feature_axis,
                                                    add_self_loops=add_self_loops,
                                                    allow_zero_in_degree=allow_zero_in_degree,
                                                    validation_score=validation_score,
                                                    early_stopping=early_stopping,
                                                    es_patience=es_patience,
                                                    es_tolerance=es_tolerance,
                                                    es_delta=es_delta,
                                                    verbose=verbose,
                                                    logs=logs,
                                                    **kwargs)
        self.in_dim = in_dim
        self.hidden_layers = hidden_layers
        self.hidden_dim = hidden_dim

    def _init_model(self, X=None, y=None):
        self.model = SGConvClassifier(self.in_dim, self.hidden_dim,
                                      len(np.unique(y)), self.hidden_layers,
                                      allow_zero_in_degree=self.allow_zero_in_degree)


class SGConvRegressorModel(DGLRegressorBaseModel):

    def __init__(self,
                 in_dim: int = 1,
                 hidden_layers: int = 2,
                 hidden_dim: int = 256,
                 nn_epochs: int = 200,
                 learning_rate: float = 0.001,
                 batch_size: int = 32,
                 adjacency_axis: int = 0,
                 feature_axis: int = 1,
                 add_self_loops: bool = True,
                 allow_zero_in_degree: bool = False,
                 validation_score: bool = False,
                 early_stopping: bool = False,
                 es_patience: int = 10,
                 es_tolerance: int = 9,
                 es_delta: float = 0,
                 verbose: bool = False,
                 logs: str = '',
                 **kwargs):
        """
        Graph convolutional network for graph regression. Simple Graph
        convolutional layers from Wu, Felix, et al., 2018. Implementation
        based on dgl & pytorch.


        Parameters
        ----------
        in_dim: int,default=1
            input dimension
        hidden_layers: int,default=2
            number of hidden layers used by the model
        hidden_dim: int,default=256
            dimensions in the hidden layers
        validation_score: bool,default=False
            If true the input data is split into train and test (90%/10%).
            The testset is then used to get validation results during training
        verbose: bool,default=False
            If true verbose output is generated
        """
        super(SGConvRegressorModel, self).__init__(nn_epochs=nn_epochs,
                                                   learning_rate=learning_rate,
                                                   batch_size=batch_size,
                                                   adjacency_axis=adjacency_axis,
                                                   feature_axis=feature_axis,
                                                   add_self_loops=add_self_loops,
                                                   allow_zero_in_degree=allow_zero_in_degree,
                                                   validation_score=validation_score,
                                                   early_stopping=early_stopping,
                                                   es_patience=es_patience,
                                                   es_tolerance=es_tolerance,
                                                   es_delta=es_delta,
                                                   verbose=verbose,
                                                   logs=logs,
                                                   **kwargs)
        self.in_dim = in_dim
        self.hidden_layers = hidden_layers
        self.hidden_dim = hidden_dim

    def _init_model(self, X=None, y=None):
        self.model = SGConvClassifier(self.in_dim, self.hidden_dim, 1, self.hidden_layers,
                                      allow_zero_in_degree=self.allow_zero_in_degree).float()



import networkx as nx
import numpy as np
import igraph
import warnings
from photonai_graph.util import assert_imported

try:
    import dgl
    import torch
except ImportError:  # pragma: no cover
    pass


def dense_to_igraph(graphs: np.ndarray, adjacency_axis: int = None, feature_axis: int = None):
    """
    Converts dense matrices to igraph Graph objects
    :param graphs:  np.ndarray or np.matrix
                    graphs represented as dense format
    :param adjacency_axis: int, default=None
                           position of the adjacency axis
    :param feature_axis: int, default=None
                         position of the feature axis
    :return: igraph_graphs: Converted list of Igraph objects
    """
    if adjacency_axis is None:
        warnings.warn("No adjacency passed. Guessing that adjacency is in first channel...")
        adjacency_axis = 0
    igraph_graphs = [igraph.Graph.Adjacency(graphs[i, ..., adjacency_axis]) for i in range(graphs.shape[0])]
    if feature_axis is not None:
        raise NotImplementedError("This feature is not implemented yet.")
    return igraph_graphs


def dense_to_networkx(graphs: np.ndarray, adjacency_axis: int = None, feature_axis=None):
    if adjacency_axis is None:
        warnings.warn("No adjacency passed. Guessing that adjacency is in first channel...")
        adjacency_axis = 0
    nx_graphs = [nx.from_numpy_array(graphs[i, ..., adjacency_axis]) for i in range(graphs.shape[0])]
    if feature_axis is not None:
        raise NotImplementedError("This feature is not implemented yet.")
    return nx_graphs


def dense_to_dgl(graphs, adjacency_axis=None, feature_axis=None):
    """Converts dense matrices to dgl graphs

        Parameters
        ---------
        graphs: list, np.ndarray or np.matrix
            graphs represented in dense format
        adjacency_axis: int, default=None
            position of the adjacency matrix
        feature_axis: int, default=None
            position of the feature matrix
    """
    assert_imported(["dgl"])

    if adjacency_axis is None:
        raise NotImplementedError('dense to dgl not implemented without adjacency axis')
    if not isinstance(graphs, np.ndarray):
        raise ValueError(f"Expected np.ndarray as input, got {type(graphs)}")

    graph_list = []
    for graph in range(graphs.shape[0]):
        src, dst = np.nonzero(graphs[graph, :, :, adjacency_axis])
        g = dgl.graph((src, dst), num_nodes=graphs[graph, ...].shape[1])
        feat = torch.tensor(graphs[graph, :, :, feature_axis])
        g.ndata['feat'] = feat
        graph_list.append(g)
    return graph_list


def sparse_to_dense(graphs, features=None):
    """convert sparse matrices to numpy array

        Parameters
        ---------
        graphs: list or scipy matrix
            a list of scipy sparse matrices or a single sparse matrix
        features: list or scipy matrix, default=None
            if a feature matrix or a list of those is specified they are
            incorporated into the numpy array
    """
    def convert_matrix(current_graph_mtrx, current_features):
        """
        Helper function
        """
        current_graph_mtrx = np.reshape(current_graph_mtrx,
                                        (current_graph_mtrx.shape[0], current_graph_mtrx.shape[1], -1))
        current_features = current_features.toarray()
        current_features = np.reshape(current_features, (current_features.shape[0], current_features.shape[1], -1))
        return np.concatenate((current_graph_mtrx, current_features), axis=2)

    if isinstance(graphs, list):
        matrices = []
        for idx, graph in enumerate(graphs):
            graph_mtrx = graph.toarray()
            if features is not None:
                matrices.append(convert_matrix(graph_mtrx, features[idx]))
            else:
                matrices.append(graph_mtrx)
    else:
        try:
            graph_mtrx = graphs.toarray()
            if features is not None:
                matrices = convert_matrix(graph_mtrx, features)
        except Exception as e:
            print('Could not convert matrices.'
                  'Your matrices need to be a list or a single sparse matrix.')
            raise e

    matrices = np.asarray(matrices)
    return matrices


def dgl_to_dense(graphs, in_fmt="csr"):
    """turns dgl graphs into dense matrices

        Parameters
        ---------
        graphs: list
            list of dgl graphs
        in_fmt: str, default="csr"
            format of the scipy sparse matrix used in the intermediary step
    """
    assert_imported(["dgl"])

    if not isinstance(graphs, list):
        raise ValueError('Input graphs need to be in list format')

    sp_graphs = dgl_to_sparse(graphs, in_fmt)
    graph_list = sparse_to_dense(sp_graphs)
    return graph_list


def dgl_to_sparse(graphs, fmt="csr"):
    """turns dgl graphs into sparse matricesParameters

        Parameters
        ---------
        graphs: list
            list of dgl graphs
        fmt: str, default="csr"
            format of the scipy sparse matrix used in the intermediary step
    """
    assert_imported(["dgl"])

    if not isinstance(graphs, list):
        raise ValueError("Input type needs to be a list")

    graph_list = []
    for graph in graphs:
        scp_graph = graph.adjacency_matrix(scipy_fmt=fmt)
        graph_list.append(scp_graph)
    return graph_list


def check_dgl(graphs, adjacency_axis=None, feature_axis=None):
    """Checks the input and converts it to dgl format

        Parameters
        ---------
        graphs: list, np.ndarray or np.matrix
            graphs to be converted
        adjacency_axis: int, default=None
            position of the adjacency matrix
        feature_axis: int, default=None
            position of the feature matrix
    """
    assert_imported(["dgl"])

    if not isinstance(graphs, np.ndarray) and not isinstance(graphs, np.matrix) and not isinstance(graphs, list):
        raise TypeError('can only handle np arrays or lists as input')

    if isinstance(graphs, list) or np.ndim(graphs) == 1:
        if not isinstance((graphs[0]), dgl.DGLGraph):
            raise ValueError("Expected list of dglGraph or dense matrix")
        dgl_graphs = graphs
    elif np.ndim(graphs) > 1:
        dgl_graphs = dense_to_dgl(graphs, adjacency_axis, feature_axis)
    else:
        raise ValueError('numpy matrix must have one or four dimensions')

    return dgl_graphs



import unittest
import numpy as np
from photonai_graph.Controllability.controllability_functions import average_control


class AverageControllabilityTests(unittest.TestCase):

    def setUp(self):
        b = np.random.randint(2, size=(20, 20))
        b_symm = (b + b.T) / 2
        b_symm[b_symm < 0.6] = 0
        self.X_sym = b_symm
        self.X_ident = np.eye(20)

    def test_ave_control(self):
        mdr = average_control(self.X_sym)
        self.assertTrue(np.all(mdr >= 0.75))

    def test_ave_control_shape(self):
        mdr = average_control(self.X_sym)
        self.assertEqual(mdr.shape, (20,))

    def test_ave_control_mechanism(self):
        mdr = average_control(self.X_ident)
        self.assertTrue(np.all(mdr == mdr[0]))



import unittest
import numpy as np
from photonai_graph.GraphConstruction.graph_constructor_random_walks import GraphConstructorRandomWalks


class RWTests(unittest.TestCase):

    def setUp(self):
        self.X4d = np.ones((20, 20, 20, 2))
        self.Xrandom4d = np.random.rand(20, 20, 20, 2)
        self.y = np.ones((20))

    def test_rw_mean(self):
        g_constr = GraphConstructorRandomWalks()
        g_constr.fit(self.Xrandom4d, self.y)
        trans = g_constr.transform(self.Xrandom4d)
        self.assertEqual(np.shape(trans), (20, 20, 20, 3))

    def test_rw_assert_failure(self):
        with self.assertRaises(AssertionError):
            g_constr = GraphConstructorRandomWalks()
            g_constr.fit(self.X4d, self.y)
            g_constr.transform(self.X4d)



try:
    import torch
    from torch.utils.data import Dataset
except ImportError:
    pass


class CustomDataset(Dataset):
    """Face Landmarks dataset."""

    def __init__(self, graphs, y):
        """
        Args:
            csv_file (string): Path to the csv file with annotations.
            root_dir (string): Directory with all the images.
            transform (callable, optional): Optional transform to be applied
                on a sample.
        """
        self.graphs = graphs
        self.labels = y

    def __len__(self):
        return len(self.graphs)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        return self.graphs[idx], self.labels[idx]


class DGLData(Dataset):
    """Dataset Class for handling dgl graphs with labels."""

    def __init__(self, samples):
        """
        Args:
            csv_file (string): Path to the csv file with annotations.
            root_dir (string): Directory with all the images.
            transform (callable, optional): Optional transform to be applied
                on a sample.
        """
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        return self.samples[idx]


def zip_data(X, y):
    # returns a zipped list from two input graphs
    data = tuple(zip(X, y))

    return data
